// ==========================================================================
// $Id$
// ==========================================================================
// (C)opyright:
//
//   Max-Planck-Institut fuer Informatik
//   Stuhlsatzenhausweg 85, 66123 Saarbruecken
//   http://www.mpi-sb.mpg.de
// 
// Creator: lensch (Hendrik Lensch)
// Email:   lensch@mpi-sb.mpg.de
// ==========================================================================
// $Log$
// ==========================================================================

#ifndef EX1_TESTGAMMA_C
#define EX1_TESTGAMMA_C

#include <stdio.h>
#include <cutil.h> 
#include <vector_types.h>

#include "PPM.hh"

using namespace std;
using namespace ppm; 

#define MAX_BLOCKS 512
#define MAX_THREADS 128

/* compute gamma correction on the float image _src of resolution dim,
   outputs the gamma corrected image should be stored in_dst[blockIdx.x *
   blockDim.x + threadIdx.x]. Each thread computes on pixel element.
*/ 
__global__ void diffKernel(float *_dst, const float* _src1, const float* _src2, int _w ) {
  
  int x = blockIdx.x * MAX_THREADS + threadIdx.x; 
  int y = blockIdx.y; 
  int pos = y * _w + x; 

   if (x < _w) 
     _dst[ pos ] = abs( _src1[ pos ] - _src2[ pos ] ); 
}

  
/* This program sets up two large arrays of size dim and computes the
   dot product of both arrays. 

   The arrays are uploaed only once and the dot product is computed
   multiple times. While this does not make too much sense it
   demonstrated the possible speedup.  */ 

int main(int argc, char* argv[]) {

  //  CUT_DEVICE_INIT( argc, argv ); 

  // parse command line 
  int acount = 1; 

  if (argc < 5) { 
    printf("usage: difference <img1> <img2> <mode> <outImg>"); 
    exit(1) ;
  }

  float* img1;  
  float* img2;

  int w1,h1;
  readPPM( argv[acount++], w1, h1, &img1 );
  int w2,h2;
  readPPM( argv[acount++], w2, h2, &img2 );
 
  if (w1!=w2 || h1!=h2)
  {
    printf("the two input images must have the same size!");
    exit(1);
  }

  // flag indicating weather the CPU or the GPU version should be executed
  bool gpuVersion = atoi( argv[acount++] ); 

  int nPix = w1*h1; 

  float* gpuImg1;
  float* gpuImg2;
  float* gpuResImg; 
  
  // copy the image to the device
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuImg1, nPix* 3 * sizeof(float) ) );
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuImg2, nPix* 3 * sizeof(float) ) );
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuResImg, nPix* 3 * sizeof(float) ) );   
  CUDA_SAFE_CALL( cudaMemcpy( gpuImg1, img1, nPix*3 * sizeof(float),  cudaMemcpyHostToDevice) );
  CUDA_SAFE_CALL( cudaMemcpy( gpuImg2, img2, nPix*3 * sizeof(float),  cudaMemcpyHostToDevice) );

  // calculate the block dimensions
  // simple idea: chop the linear area of nPix*3 samples into
  // MAX_THREADS: 

  dim3 threadBlock( MAX_THREADS, 1 ); 
  // select a random number of bocks vertically 
  
  dim3 blockGrid( (w1*3)/MAX_THREADS +1, h1,  1); 

  printf( "bl/thr: %d  %d \n", blockGrid.x, threadBlock.x ); 

  if (gpuVersion) {
    diffKernel<<< blockGrid, threadBlock >>>( gpuResImg, gpuImg1, gpuImg2, w1 * 3); 
     CUT_CHECK_ERROR("diff failed\n");
     CUDA_SAFE_CALL( cudaThreadSynchronize() );	
     // download result
     CUDA_SAFE_CALL( cudaMemcpy( img1, gpuResImg, nPix*3 * sizeof(float),  cudaMemcpyDeviceToHost) );
     CUT_CHECK_ERROR("copy failed\n");

  } else {
    for (int i = 0; i < nPix * 3; ++i) {
      img1[i] = abs(img1[i] - img2[i]); 
    }
  }


  cudaFree( gpuResImg ); 
  cudaFree( gpuImg1 );
  cudaFree( gpuImg2 );

  writePPM( argv[acount++], w1, h1, (float*)img1 ); 

  delete[] img1; 
  delete[] img2;

  printf("done\n"); 

}

    

      
      

      
   

      

      
      





#endif /* EX1_TESTGAMMA_C */







