// ==========================================================================
// $id$++-
// ==========================================================================
// (C)opyright:
//
//   Max-Planck-Institut fuer Informatik
//   Stuhlsatzenhausweg 85, 66123 Saarbruecken
//   http://www.mpi-sb.mpg.de
// 
// Creator: lensch (Hendrik Lensch)
// Email:   lensch@mpi-sb.mpg.de
// ==========================================================================
// $Log$
// ==========================================================================

#ifndef EX1_TESTGAMMA_C
#define EX1_TESTGAMMA_C

#include <stdio.h>
#include <cutil.h> 
#include <vector_types.h>

#include "PPM.hh"

using namespace std;
using namespace ppm; 

#define MAX_BLOCKS 512000
#define MAX_THREADS 32

// specify the gamma value to be applied
__device__ __constant__ float gpuGamma[2]; 

float applyGammaHost( const float& _src, const float _gamma ) {
  
  return 255. * powf( _src / 255., _gamma); 

}
  

__device__ float applyGamma( const float& _src, const float _gamma ) {
  
  return 255. * __powf( _src / 255., _gamma); 

}
  
 
/* compute gamma correction on the float image _src of resolution dim,
   outputs the gamma corrected image should be stored in_dst[blockIdx.x *
   blockDim.x + threadIdx.x]. Each thread computes on pixel element.
*/ 
__global__ void gammaKernel(float *_dst, const float* _src, int
   _w ) {
  
  int x = blockIdx.x * MAX_THREADS + threadIdx.x; 
  int y = blockIdx.y; 
  int pos = y * _w + x; 

   if (x < _w) 
     _dst[ pos ] = applyGamma( _src[ pos ], gpuGamma[0]); 
   
}

  
/* This program sets up two large arrays of size dim and computes the
   dot product of both arrays. 

   The arrays are uploaed only once and the dot product is computed
   multiple times. While this does not make too much sense it
   demonstrated the possible speedup.  */ 

int main(int argc, char* argv[]) {

  // parse command line 
  int acount = 1; 

  if (argc < 5) { 
    printf("usage: testGamma <inImg> <gamma> <mode> <outImg>"); 
    exit(1) ;
  }

  float* img;  

  int w,h;
  readPPM( argv[acount++], w, h, &img ); 
 
  float gamma = atof( argv[acount++] ); 
  // flag indicating weather the CPU or the GPU version should be executed
  bool gpuVersion = atoi( argv[acount++] ); 

  int nPix = w*h; 

  float* gpuImg; 
  float* gpuResImg; 
  
  // copy the image to the device
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuImg, nPix* 3 * sizeof(float) ) );   
  CUDA_SAFE_CALL( cudaMalloc( (void**)&gpuResImg, nPix* 3 * sizeof(float) ) );   
  CUDA_SAFE_CALL( cudaMemcpy( gpuImg, img, nPix*3 * sizeof(float),  cudaMemcpyHostToDevice) );

  //  CUDA_SAFE_CALL( cudaMemcpy( gpuResImg, img, nPix*3 * sizeof(float),  cudaMemcpyHostToDevice) );
    // copy gamma value to constant device memory
  CUDA_SAFE_CALL( cudaMemcpyToSymbol(gpuGamma, &gamma, sizeof(float)) );

  // calculate the block dimensions
  // simple idea: chop the linear area of nPix*3 samples into
  // MAX_THREADS: 

  dim3 threadBlock( MAX_THREADS, 1 ); 
  // select a random number of bocks vertically 
  
  dim3 blockGrid( (w*3)/MAX_THREADS , h,  1); 

  printf( "gamma: %f\n", gamma ); 
  printf( "bl/thr: %d  %d \n", blockGrid.x, threadBlock.x ); 

  if (gpuVersion) {
    gammaKernel<<< blockGrid, threadBlock >>>( gpuResImg, gpuImg, w * 3); 
     CUT_CHECK_ERROR("gamma filter failed\n");
     CUDA_SAFE_CALL( cudaThreadSynchronize() );	
     // download result
     CUDA_SAFE_CALL( cudaMemcpy( img, gpuResImg, nPix*3 * sizeof(float),  cudaMemcpyDeviceToHost) );
     CUT_CHECK_ERROR("copy failed\n");

  } else {
    for (int i = 0; i < nPix * 3; ++i) {
      img[i] = applyGammaHost( img[i], gamma ); 
    }
  }


  cudaFree( gpuResImg ); 
  cudaFree( gpuImg ); 

  writePPM( argv[acount++], w, h, (float*)img ); 

  delete[] img; 

  printf("done\n"); 

}

    

      
      

      
   

      

      
      





#endif /* EX1_TESTGAMMA_C */







