
#include <cutil.h>

#include <stdio.h>

#include "CudaArray.h"

#include <vector>

float transmitterPower = 300.f; // Transmission power, as measured 0.001 units away from the transmitter
float minimumSignalPower = 1.f; // Lowest possible signal power under which the receiver can pick up the signal

enum { BucketSize = 1024 };

enum { BucketsPerAxis = 16 };


struct Transmitter
{
  float2 position;
  float powerScale;
  int pad;          // Structure must be a multiple of 8 bytes large, due to usage of "float2" datatype
};

struct Receiver
{
  float2 position;
  float minimumSignalPower;
  int pad;          // Structure must be a multiple of 8 bytes large, due to usage of "float2" datatype
};

struct TransmitterBucket
{
  unsigned int numTransmitters;
  const Transmitter* transmitters[BucketSize];
};

typedef std::vector<Transmitter> TransmitterList;
typedef std::vector<Receiver> ReceiverList;

typedef std::vector<float2> PositionList;

typedef std::vector<float> SignalStrengthList;


///////////////////////////////////////////////////////////////////////////////////////////////
// Given a transmitter and a receiver, calculate signal strength
//
// The signal strength is proportional to 1/distance^2
// When the signal strength goes beneath the receiver's minimum acceptable strength,
//  the strength returned will be zero.
//
// The __host__ __device__ tags allows it to be called both from CPU and from CUDA code

__host__ __device__ float signalStrength(const Transmitter& transmitter,
					 const Receiver& receiver)
{
  // Calculate distance between transmitter and receiver

  float2 positionDelta;
  positionDelta.x = receiver.position.x - transmitter.position.x;
  positionDelta.y = receiver.position.y - transmitter.position.y;

  float distance = sqrt(positionDelta.x * positionDelta.x
			+ positionDelta.y * positionDelta.y);

  // Ensure that distance never is zero; otherwise the 1/distance^2 calculation would
  //  divide-by-zero

  if (distance < 1e-10)
    distance = 1e-10;

  // Calculate signal power at the receiver's location
 
  float normalizedPower = 1.f / (distance * distance);

  float strength = transmitter.powerScale * normalizedPower;

  // Clamp too low signal power to zero

  if (strength < receiver.minimumSignalPower)
    strength = 0.f;

  return strength;
}


///////////////////////////////////////////////////////////////////////////////////////////////


PositionList generateRandomPositions(int numPositions)
{
  PositionList positions;

  for (int i = 0; i < numPositions; ++i)
    {
      float2 position;
      position.x = ((float) rand()) / RAND_MAX;
      position.y = ((float) rand()) / RAND_MAX;
      positions.push_back(position);
    }

  return positions;
}

TransmitterList createTransmitters(int numTransmitters)
{
  PositionList positions = generateRandomPositions(numTransmitters);

  TransmitterList transmitters;

  for (int i = 0; i < numTransmitters; ++i)
    {
      Transmitter transmitter;

      transmitter.position = positions[i];
      transmitter.powerScale = transmitterPower * 0.001f * 0.001f;

      transmitters.push_back(transmitter);
    }

  return transmitters;
}

ReceiverList createReceivers(int numReceivers)
{
  PositionList positions = generateRandomPositions(numReceivers);

  ReceiverList receivers;

  for (int i = 0; i < numReceivers; ++i)
    {
      Receiver receiver;
      receiver.position = positions[i];
      receiver.minimumSignalPower = minimumSignalPower;
      receivers.push_back(receiver);
    }

  return receivers;
}

///////////////////////////////////////////////////////////////////////////////////////////////

void calcSignalStrengthStatistics(const SignalStrengthList& strengths,
				    float& minStrength, float& maxStrength,
				    float& averageStrength,
				    int& numReceiversWithoutCoverage)
{
  if (strengths.empty())
    {
      minStrength = 0.f;
      maxStrength = 0.f;
      averageStrength = 0.f;
      numReceiversWithoutCoverage = 0;
    }
  else
    {
      minStrength = 1e20;
      maxStrength = 0.f;
      averageStrength = 0.f;
      numReceiversWithoutCoverage = 0;

      for (SignalStrengthList::const_iterator strength = strengths.begin();
	   strength != strengths.end();
	   ++strength)
	{
	  if (minStrength > *strength)
	    minStrength = *strength;
	  if (maxStrength < *strength)
	    maxStrength = *strength;

	  averageStrength += *strength;

	  if (*strength < 0.00001f)
	    numReceiversWithoutCoverage++;
	}

      averageStrength /= strengths.size();
    }
}


///////////////////////////////////////////////////////////////////////////////////////////////
// Brute force CUDA implementation which computes signal strengths

__global__ void calculateSignalStrengthsKernel(const Transmitter* transmitters,
					       int numTransmitters,
					       const Receiver* receivers,
					       int numReceivers,
					       float* signalStrengths)
{
  int receiverIndex = blockIdx.x * blockDim.x + threadIdx.x;

  if (receiverIndex < numReceivers)
    {
      const Receiver* receiver = &receivers[receiverIndex];
      float* finalStrength = &signalStrengths[receiverIndex];

      float bestSignalStrength = 0.f;

      for (int transmitterIndex = 0; transmitterIndex < numTransmitters; ++transmitterIndex)
	{
	  const Transmitter* transmitter = &transmitters[transmitterIndex];

	  float strength = signalStrength(*transmitter, *receiver);
	  
	  if (bestSignalStrength < strength)
	    bestSignalStrength = strength;
	}

      *finalStrength = bestSignalStrength;
    }
}


void calculateSignalStrengthsCuda(const TransmitterList& cpuTransmitters,
				  const ReceiverList& cpuReceivers,
				  SignalStrengthList& cpuSignalStrengths)
{
  // Allocate device memory for input and output arrays

  CudaArray<Transmitter> cudaTransmitters(cpuTransmitters.size());
  CudaArray<Receiver> cudaReceivers(cpuReceivers.size());
  CudaArray<float> cudaSignalStrengths(cpuReceivers.size());

  // Copy transmitter & receiver arrays to device memory

  cudaTransmitters.copyToCuda(&(*cpuTransmitters.begin()));
  cudaReceivers.copyToCuda(&(*cpuReceivers.begin()));

  // Perform signal strength computation

  int numThreads = std::min(cudaReceivers.size(), 256);
  int numBlocks = (cudaReceivers.size() + numThreads - 1) / numThreads;

  calculateSignalStrengthsKernel<<< numBlocks, numThreads >>>(cudaTransmitters.cudaArray(),
							      cudaTransmitters.size(),
							      cudaReceivers.cudaArray(),
							      cudaReceivers.size(),
							      cudaSignalStrengths.cudaArray());
  CUDA_SAFE_CALL(cudaThreadSynchronize());
  CUT_CHECK_ERROR("Kernel execution failed");

  // Copy results back to host memory

  cpuSignalStrengths.resize(cudaSignalStrengths.size());
  cudaSignalStrengths.copyFromCuda(&(*cpuSignalStrengths.begin()));
}

///////////////////////////////////////////////////////////////////////////////////////////////
// "Smart" CUDA implementation which computes signal strengths
//
// First, all transmitters are stored into buckets
// Then, receivers only compute signal strength against transmitters in nearby buckets


// Perform bucket-sort on transmitters

__global__ void sortTransmittersIntoBucketsKernel(const Transmitter* transmitters,
					    int numTransmitters,
					    TransmitterBucket* buckets)
{
  int transmitterIndex = blockIdx.x * blockDim.x + threadIdx.x;

  if (transmitterIndex < numTransmitters)
    {
      const Transmitter* transmitter = &transmitters[transmitterIndex];

      // Calculate bucket index for transmitter

      int bucketIndexX = (int) (transmitter->position.x * BucketsPerAxis);
      int bucketIndexY = (int) (transmitter->position.y * BucketsPerAxis);

      // Clamp bucket index to [0, BucketsPerAxis - 1] along both axes

      if (bucketIndexX < 0)
	bucketIndexX = 0;
      if (bucketIndexX >= BucketsPerAxis)
	bucketIndexX = BucketsPerAxis - 1;

      if (bucketIndexY < 0)
	bucketIndexY = 0;
      if (bucketIndexY >= BucketsPerAxis)
	bucketIndexY = BucketsPerAxis - 1;

      int bucketIndex = bucketIndexY * BucketsPerAxis + bucketIndexX;

      // Store transmitter into bucket

      TransmitterBucket* bucket = &buckets[bucketIndex];

      // Use atomic increase operation to allocate one slot in bucket,
      //  without risking any race conditions

      int offset = (int) atomicInc(&bucket->numTransmitters, 0xffffffffU);

      bucket->transmitters[offset] = transmitter;
    }
}

// Go through one bucket, find highest signal strength for all transmitters in it
// Return highest strength (or the old value, if it was higher)

__device__ float scanBucket(const TransmitterBucket* transmitterBucket, const Receiver* receiver, float bestSignalStrength)
{
  int numTransmitters = (int) transmitterBucket->numTransmitters;

  for (int transmitterIndex = 0; transmitterIndex < numTransmitters; ++transmitterIndex)
    {
      const Transmitter* transmitter = transmitterBucket->transmitters[transmitterIndex];

      float strength = signalStrength(*transmitter, *receiver);

      if (bestSignalStrength < strength)
	bestSignalStrength = strength;
    }

  return bestSignalStrength;
}

// Calculate signal strength for all receivers

__global__ void calculateSignalStrengthsKernel2(const TransmitterBucket* transmitterBuckets,
					       const Receiver* receivers,
					       int numReceivers,
					       float* signalStrengths)
{
  int receiverIndex = blockIdx.x * blockDim.x + threadIdx.x;

  if (receiverIndex < numReceivers)
    {
      const Receiver* receiver = &receivers[receiverIndex];
      float* finalStrength = &signalStrengths[receiverIndex];

      // Compute bucket index for receiver

      int receiverBucketIndexX = (int) (receiver->position.x * BucketsPerAxis);
      int receiverBucketIndexY = (int) (receiver->position.y * BucketsPerAxis);

      float bestSignalStrength = 0.f;

      // Scan all buckets in the 3x3 region enclosing the receiver's bucket index

      for (int bucketIndexY = receiverBucketIndexY - 1; bucketIndexY < receiverBucketIndexY + 2; ++bucketIndexY)
	for (int bucketIndexX = receiverBucketIndexX - 1; bucketIndexX < receiverBucketIndexX + 2; ++bucketIndexX)
	  {
	    // Only process bucket if its index is within [0, BucketsPerAxis - 1] along each axis

	    if (bucketIndexX >= 0 && bucketIndexX < BucketsPerAxis
		&& bucketIndexY >= 0 && bucketIndexY < BucketsPerAxis)
	      {
		// Scan bucket for a potential new "highest signal strength"

		int bucketIndex = bucketIndexY * BucketsPerAxis + bucketIndexX;
		bestSignalStrength = scanBucket(&transmitterBuckets[bucketIndex], receiver, bestSignalStrength);
	      }
	  }

      // Store out the highest signal strength found for the receiver

      *finalStrength = bestSignalStrength;
    }
}


void calculateSignalStrengthsCuda2(const TransmitterList& cpuTransmitters,
				  const ReceiverList& cpuReceivers,
				  SignalStrengthList& cpuSignalStrengths)
{
  // Allocate device memory for input and output arrays

  CudaArray<Transmitter> cudaTransmitters(cpuTransmitters.size());
  CudaArray<Receiver> cudaReceivers(cpuReceivers.size());

  // Copy transmitter & receiver arrays to device memory

  cudaTransmitters.copyToCuda(&(*cpuTransmitters.begin()));
  cudaReceivers.copyToCuda(&(*cpuReceivers.begin()));

  // Create buckets in device memory

  int numBuckets = BucketsPerAxis * BucketsPerAxis;
  CudaArray<TransmitterBucket> cudaTransmitterBuckets(numBuckets);

  // Sort transmitters into buckets

  {
    int numThreads = std::min(cudaTransmitters.size(), 256);
    int numBlocks = (cudaTransmitters.size() + numThreads - 1) / numThreads;

    cudaMemset(cudaTransmitterBuckets.cudaArray(), 0, numBuckets * sizeof(TransmitterBucket));

    sortTransmittersIntoBucketsKernel<<< numBlocks, numThreads >>>(cudaTransmitters.cudaArray(),
								   cudaTransmitters.size(),
								   cudaTransmitterBuckets.cudaArray());
    CUDA_SAFE_CALL(cudaThreadSynchronize());
    CUT_CHECK_ERROR("Kernel execution failed");
  }

  // Perform signal strength computation

  CudaArray<float> cudaSignalStrengths(cpuReceivers.size());

  {
    int numThreads = std::min(cudaReceivers.size(), 256);
    int numBlocks = (cudaReceivers.size() + numThreads - 1) / numThreads;

    calculateSignalStrengthsKernel2<<< numBlocks, numThreads >>>(cudaTransmitterBuckets.cudaArray(),
								 cudaReceivers.cudaArray(),
								 cudaReceivers.size(),
								 cudaSignalStrengths.cudaArray());
    CUDA_SAFE_CALL(cudaThreadSynchronize());
    CUT_CHECK_ERROR("Kernel execution failed");
  }

  // Copy results back to host memory

  cpuSignalStrengths.resize(cudaSignalStrengths.size());
  cudaSignalStrengths.copyFromCuda(&(*cpuSignalStrengths.begin()));
}


///////////////////////////////////////////////////////////////////////////////////////////////
// Brute force CPU implementation which calculates signal strengths
//

void calculateSignalStrengthsCpu(const TransmitterList& transmitters,
				 const ReceiverList& receivers,
				 SignalStrengthList& signalStrengths)
{
  signalStrengths.clear();

  for (ReceiverList::const_iterator receiver = receivers.begin();
       receiver != receivers.end();
       ++receiver)
    {
      
      float bestSignalStrength = 0.f;
      
      for (TransmitterList::const_iterator transmitter = transmitters.begin();
	   transmitter != transmitters.end();
	   ++transmitter)
	{
	  float strength = signalStrength(*transmitter, *receiver);
	  
	  if (bestSignalStrength < strength)
	    bestSignalStrength = strength;
	}
      
      signalStrengths.push_back(bestSignalStrength);
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////


int main(int argc, char** argv)
{
  if (argc != 4)
    {
      printf("Usage: CellPhoneCoverage <num transmitters> <num receivers> <method [0, 1, 2]>\n");
      return 1;
    }

  int numTransmitters = atoi(argv[1]);
  int numReceivers = atoi(argv[2]);

  int method = atoi(argv[3]);

  float transmissionRadius = sqrt((transmitterPower * 0.001f * 0.001f) / minimumSignalPower);
  float bucketSideLength = 1.f / BucketsPerAxis;

  if (transmissionRadius > bucketSideLength / 2)
    {
      printf("Error: Transmission radius is %f units, but must not be greater than %f units\n", transmissionRadius, bucketSideLength / 2);
      return 1;
    }


  CUT_DEVICE_INIT(argc, argv);

  printf("\n");

  TransmitterList cpuTransmitters = createTransmitters(numTransmitters);
  ReceiverList cpuReceivers = createReceivers(numReceivers);

  SignalStrengthList cpuSignalStrengths;

  switch (method)
    {
    case 0:
      printf("Performing brute-force CPU computation\n");
      calculateSignalStrengthsCpu(cpuTransmitters, cpuReceivers, cpuSignalStrengths);
      break;
    case 1:
      printf("Performing brute-force CUDA computation\n");
      calculateSignalStrengthsCuda(cpuTransmitters, cpuReceivers, cpuSignalStrengths);
      break;
    case 2:
      printf("Performing sorted CUDA computation\n");
      calculateSignalStrengthsCuda2(cpuTransmitters, cpuReceivers, cpuSignalStrengths);
      break;
    }

  float minStrength;
  float maxStrength;
  float averageStrength;
  int numReceiversWithoutCoverage;

  calcSignalStrengthStatistics(cpuSignalStrengths,
			       minStrength, maxStrength, averageStrength,
			       numReceiversWithoutCoverage);

  printf("Number of transmitters: %d\n", numTransmitters);
  printf("Number of receivers: %d\n", numReceivers);
  printf("Transmission radius for each transmitter: %f units\n", transmissionRadius);
  printf("Min strength: %f W\n", minStrength);
  printf("Max strength: %f W\n", maxStrength);
  printf("Average strength: %f W\n", averageStrength);
  printf("Number of receivers without coverage: %d (%.1f%%)\n", numReceiversWithoutCoverage,
	 100.f * numReceiversWithoutCoverage / numReceivers);

  CUT_EXIT(argc, argv);
 
  return 0;
}
