
#include "SignalStrengthsSortedCuda.h"

#include "CellPhoneCoverage.h"
#include "CudaArray.h"
#include "Helpers.h"


// "Smart" CUDA implementation which computes signal strengths
//
// First, all transmitters are sorted into buckets
// Then, all receivers are sorted into buckets
// Then, receivers only compute signal strength against transmitters in nearby buckets
//
// This multi-step algorithm makes the signal strength computation scale much
//  better to high number of transmitters/receivers



struct Bucket
{
  int startIndex;         // Start of bucket within array
  int numElements;        // Number of elements in bucket
};


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Given a set of positions, compute how many of these will land in the respective bucket
//

static __global__ void buildBucketHistogram(const Position* positions,
				     int numPositions,
				     unsigned int* histogram)
{
  int positionIndex = blockIdx.x * blockDim.x + threadIdx.x;

  if (positionIndex < numPositions)
    {

      const Position& position = positions[positionIndex];

      int bucketIndexX = (int) (position.x * BucketsPerAxis);
      int bucketIndexY = (int) (position.y * BucketsPerAxis);

      int bucketIndex = bucketIndexY * BucketsPerAxis + bucketIndexX;

      atomicInc(&histogram[bucketIndex], 0xffffffffU);
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Convert histogram into an array of (starting index, number of element) pairs

static __global__ void convertHistogramToBuckets(const unsigned int* histogram,
					  int numElements,
					  Bucket* buckets)
{
  buckets[0].startIndex = 0;
  buckets[0].numElements = (int) histogram[0];
  for (int i = 0; i < numElements - 1; ++i)
    {
      buckets[i + 1].startIndex = buckets[i].startIndex + buckets[i].numElements;
      buckets[i + 1].numElements = (int) histogram[i + 1];
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Copy out the bucket start indices to a new array

static __global__ void setupBucketWriteOffsets(const Bucket* buckets,
					int numElements,
					unsigned int* writeOffsets)
{
  for (int i = 0; i < numElements; ++i)
    writeOffsets[i] = (unsigned int) buckets[i].startIndex;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Given the start-indices, copy 

static __global__ void remapPositions(const Position* inputPositions,
			       int numPositions,
			       Position* outputPositions,
			       unsigned int* writeOffsets)
{
  int inputIndex = blockIdx.x * blockDim.x + threadIdx.x;

  if (inputIndex < numPositions)
    {
      const Position& position = inputPositions[inputIndex];

      int bucketIndexX = (int) (position.x * BucketsPerAxis);
      int bucketIndexY = (int) (position.y * BucketsPerAxis);

      int bucketIndex = bucketIndexY * BucketsPerAxis + bucketIndexX;

      int outputIndex = (int) atomicInc(&writeOffsets[bucketIndex], 0xffffffffU);

      outputPositions[outputIndex] = position;
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Sort a set of positions into a set of buckets
//
// Given a set of input positions, these will be re-ordered such that
//  each range of elements in the output array belong to the same bucket.
// The list of buckets that is output describes where each such range begins
//  and ends in the re-ordered position array.

static void sortPositionsIntoBuckets(CudaArray<Position>& cudaInputPositions,
			      CudaArray<Position>& cudaOutputPositions,
			      CudaArray<Bucket>& cudaOutputPositionBuckets)
{
  // Insert counting, phase 1:

  // Build histogram for positions
  
  CudaArray<unsigned int> bucketHistogram(cudaOutputPositionBuckets.size());

  int numThreads = std::min(cudaInputPositions.size(), 256);
  int numBlocks = (cudaInputPositions.size() + numThreads - 1) / numThreads;

  cudaMemset(bucketHistogram.cudaArray(), 0, bucketHistogram.size() *
	     sizeof(unsigned int));

  buildBucketHistogram<<< numBlocks, numThreads >>>(cudaInputPositions.cudaArray(),
						    cudaInputPositions.size(),
						    bucketHistogram.cudaArray());
  CUDA_SAFE_CALL(cudaThreadSynchronize());
  CUT_CHECK_ERROR("Kernel execution failed");

  // Insert counting, phase 2:

  // Convert histogram to actual set of buckets

  convertHistogramToBuckets<<< 1, 1 >>>(bucketHistogram.cudaArray(),
					bucketHistogram.size(),
					cudaOutputPositionBuckets.cudaArray());
  CUDA_SAFE_CALL(cudaThreadSynchronize());
  CUT_CHECK_ERROR("Kernel execution failed");

  // Copy bucket starting positions to a set of temporary write-offsets

  CudaArray<unsigned int> writeOffsets(cudaOutputPositionBuckets.size());

  setupBucketWriteOffsets<<< 1, 1 >>>(cudaOutputPositionBuckets.cudaArray(),
				      cudaOutputPositionBuckets.size(),
				      writeOffsets.cudaArray());
  CUDA_SAFE_CALL(cudaThreadSynchronize());
  CUT_CHECK_ERROR("Kernel execution failed");


  // Insert counting, phase 3:

  // Remap positions from temp to final array

  remapPositions<<< numBlocks, numThreads >>>(cudaInputPositions.cudaArray(),
					      cudaInputPositions.size(),
					      cudaOutputPositions.cudaArray(),
					      writeOffsets.cudaArray());
  CUDA_SAFE_CALL(cudaThreadSynchronize());
  CUT_CHECK_ERROR("Kernel execution failed");
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Go through all transmitters in one bucket, find highest signal strength
// Return highest strength (or the old value, if that was higher)

static __device__ float scanBucket(const Position* transmitters, int numTransmitters, const Position& receiver, float bestSignalStrength)
{
  for (int transmitterIndex = 0; transmitterIndex < numTransmitters; ++transmitterIndex)
    {
      const Position& transmitter = transmitters[transmitterIndex];

      float strength = signalStrength(transmitter, receiver);

      if (bestSignalStrength < strength)
	bestSignalStrength = strength;
    }

  return bestSignalStrength;
}

///////////////////////////////////////////////////////////////////////////////////////////////
//
// Calculate signal strength for all receivers

static __global__ void calculateSignalStrengthsSortedKernel(const Position* transmitters,
						     const Bucket* transmitterBuckets,
						     const Position* receivers,
						     const Bucket* receiverBuckets,
						     float* signalStrengths)
{
  // Determine which bucket the current grid block is processing

  int receiverBucketIndexX = blockIdx.x;
  int receiverBucketIndexY = blockIdx.y;

  int receiverBucketIndex = receiverBucketIndexY * BucketsPerAxis + receiverBucketIndexX;

  const Bucket& receiverBucket = receiverBuckets[receiverBucketIndex];

  int receiverStartIndex = receiverBucket.startIndex;
  int numReceivers = receiverBucket.numElements;

  // Distribute available receivers over the set of available threads

  for (int receiverIndex = threadIdx.x; receiverIndex < numReceivers; receiverIndex += blockDim.x)
    {
      // Locate current receiver within the current bucket

      const Position& receiver = receivers[receiverStartIndex + receiverIndex];
      float& finalStrength = signalStrengths[receiverStartIndex + receiverIndex];

      float bestSignalStrength = 0.f;

      // Scan all buckets in the 3x3 region enclosing the receiver's bucket index
      
      for (int transmitterBucketIndexY = receiverBucketIndexY - 1; transmitterBucketIndexY < receiverBucketIndexY + 2; ++transmitterBucketIndexY)
	for (int transmitterBucketIndexX = receiverBucketIndexX - 1; transmitterBucketIndexX < receiverBucketIndexX + 2; ++transmitterBucketIndexX)
	  {
	    // Only process bucket if its index is within [0, BucketsPerAxis - 1] along each axis

	    if (transmitterBucketIndexX >= 0 && transmitterBucketIndexX < BucketsPerAxis
		&& transmitterBucketIndexY >= 0 && transmitterBucketIndexY < BucketsPerAxis)
	      {
		// Scan bucket for a potential new "highest signal strength"

		int transmitterBucketIndex = transmitterBucketIndexY * BucketsPerAxis + transmitterBucketIndexX;
		int transmitterStartIndex = transmitterBuckets[transmitterBucketIndex].startIndex;
		int numTransmitters = transmitterBuckets[transmitterBucketIndex].numElements;
		bestSignalStrength = scanBucket(&transmitters[transmitterStartIndex], numTransmitters, receiver, bestSignalStrength);
	      }
	  }

      // Store out the highest signal strength found for the receiver

      finalStrength = bestSignalStrength;
    }
}

///////////////////////////////////////////////////////////////////////////////////////////////

void calculateSignalStrengthsSortedCuda(const PositionList& cpuTransmitters,
				  const PositionList& cpuReceivers,
				  SignalStrengthList& cpuSignalStrengths)
{
  int numBuckets = BucketsPerAxis * BucketsPerAxis;

  // Copy input positions to device memory

  CudaArray<Position> cudaTempTransmitters(cpuTransmitters.size());
  cudaTempTransmitters.copyToCuda(&(*cpuTransmitters.begin()));

  CudaArray<Position> cudaTempReceivers(cpuReceivers.size());
  cudaTempReceivers.copyToCuda(&(*cpuReceivers.begin()));

  // Allocate device memory for sorted arrays

  CudaArray<Position> cudaTransmitters(cpuTransmitters.size());
  CudaArray<Bucket> cudaTransmitterBuckets(numBuckets);

  CudaArray<Position> cudaReceivers(cpuReceivers.size());
  CudaArray<Bucket> cudaReceiverBuckets(numBuckets);

  // Sort transmitters and receivers into buckets

  sortPositionsIntoBuckets(cudaTempTransmitters, cudaTransmitters, cudaTransmitterBuckets);
  sortPositionsIntoBuckets(cudaTempReceivers, cudaReceivers, cudaReceiverBuckets);

  // Perform signal strength computation
  CudaArray<float> cudaSignalStrengths(cpuReceivers.size());

  int numThreads = 256;
  dim3 grid = dim3(BucketsPerAxis, BucketsPerAxis);

  calculateSignalStrengthsSortedKernel<<< grid, numThreads >>>(cudaTransmitters.cudaArray(),
							       cudaTransmitterBuckets.cudaArray(),
							       cudaReceivers.cudaArray(),
							       cudaReceiverBuckets.cudaArray(),
							       cudaSignalStrengths.cudaArray());
  CUDA_SAFE_CALL(cudaThreadSynchronize());
  CUT_CHECK_ERROR("Kernel execution failed");

  // Copy results back to host memory

  cpuSignalStrengths.resize(cudaSignalStrengths.size());
  cudaSignalStrengths.copyFromCuda(&(*cpuSignalStrengths.begin()));
}


